#include <algorithm>
#include <cmath>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

// U primerima se tipovi zavrsavaju sufiksom _t
// sto obicno necemo raditi u realnom softveru
// da bude ocigledno sta su nasi tipovi, a sta ne.

// kompozitna klasa (kompletna, sa virtualnim kopiranjem)

class figure_t;

class figure_observer_t
{
public:
    virtual void update(figure_t *who) = 0;

    // Ako imamo virtuelnu funkciju, obavezno je napraviti
    // i virtuelni destruktor
    virtual ~figure_observer_t()
    {
    }

    figure_observer_t()
    {
    }

private:
    figure_observer_t(const figure_observer_t &) = delete;
    figure_observer_t &operator=(const figure_observer_t &) = delete;
};

class point_t
{
public:
    point_t(double x, double y)
        : x(x)
        , y(y)
    {
    }

    ~point_t()
    {
        std::cout << "~point_t " << *this << std::endl;
    }

    double x;
    double y;

    friend std::ostream &operator<<(std::ostream &ostr, const point_t &t)
    {
        ostr << "(" << t.x << "," << t.y << ")";
        return ostr;
    }

    friend std::istream &operator>>(std::istream &istr, point_t &a)
    {
        char c1, c2, c3;
        istr >> c1 >> a.x >> c2 >> a.y >> c3;
        if (c1 != '(' || c2 != ',' || c3 != ')')
            istr.setstate(std::ios::failbit);
        return istr;
    }
};

class figure_t
{
public:
    figure_t(double x, double y)
        : m_position(x, y)
    {
        s_figure_count++;
    }

    figure_t(const figure_t &figure)
        : m_position(figure.position())
    {
        s_figure_count++;
    }

    virtual ~figure_t()
    {
        std::cout << "~figure_t" << std::endl;
        s_figure_count--;
    }

    virtual std::unique_ptr<figure_t> clone() const = 0;

    point_t position() const
    {
        return m_position;
    }

    void set_position(point_t new_position)
    {
        m_position = new_position;
        for (auto *observer : m_observers) {
            observer->update(this);
        }
    }

    virtual double area() const = 0;
    virtual std::string_view class_name() const = 0;

    virtual void print(std::ostream &ostr) const
    {
        ostr << class_name() << position() << " : P=" << area();
    }

    static std::size_t s_figure_count;

    void add_observer(figure_observer_t *observer)
    {
        m_observers.push_back(observer);
    }

    bool remove_observer(figure_observer_t *observer)
    {
        if (auto it = std::find(m_observers.begin(), m_observers.end(), observer); it != m_observers.end()) {
            m_observers.erase(it);
            return true;
        }

        return false;
    }

protected:
    point_t m_position;

private:
    std::vector<figure_observer_t *> m_observers;
};

std::size_t figure_t::s_figure_count = 0;

std::ostream &operator<<(std::ostream &ostr, const figure_t &figure)
{
    figure.print(ostr);
    return ostr;
}

class rectangle_t : public figure_t
{
public:
    rectangle_t(double x, double y, double s, double v)
        : figure_t(x, y)
        , m_width(s)
        , m_height(v)
    {
    }

    ~rectangle_t()
    {
        std::cout << "~rectangle_t" << std::endl;
    }

    std::unique_ptr<figure_t> clone() const override
    {
        return std::make_unique<rectangle_t>(*this);
    }

    double width() const
    {
        return m_width;
    }

    double height() const
    {
        return m_height;
    }

    std::string_view class_name() const override
    {
        return "rectangle_t";
    }

    double area() const override
    {
        return width() * height();
    }

    void print(std::ostream &ostr) const override
    {
        figure_t::print(ostr);
        ostr << " : a,b=" << width() << "," << height();
    }

private:
    double m_width;
    double m_height;
};

class square_t : public rectangle_t
{
public:
    square_t(double x, double y, double a)
        : rectangle_t(x, y, a, a)
    {
    }

    ~square_t()
    {
        std::cout << "~rectangle_t" << std::endl;
    }

    std::unique_ptr<figure_t> clone() const override
    {
        return std::make_unique<square_t>(*this);
    }

    std::string_view class_name() const override
    {
        return "square_t";
    }
};

class circle_t : public figure_t
{
public:
    circle_t(double x, double y, double radius)
        : figure_t(x, y)
        , m_radius(radius)
    {
    }

    ~circle_t()
    {
        std::cout << "~circle_t" << std::endl;
    }

    std::unique_ptr<figure_t> clone() const override
    {
        return std::make_unique<circle_t>(*this);
    }

    double radius() const
    {
        return m_radius;
    }

    std::string_view class_name() const override
    {
        return "circle_t";
    }
    double area() const override
    {
        return radius() * radius() * M_PI;
    }

    void print(std::ostream &ostr) const override
    {
        figure_t::print(ostr);
        ostr << " : radius=" << radius();
    }

private:
    double m_radius;
};

class composite_figure_t : public figure_t
{
public:
    composite_figure_t(double x, double y)
        : figure_t(x, y)
    {
    }

    composite_figure_t(const composite_figure_t &other)
        : figure_t(other.position().x, other.position().y)
    {
        for (const auto &figure : other.m_figures)
            add(figure->clone());
    }

    ~composite_figure_t()
    {
        std::cout << "~composite_figure_t" << std::endl;
    }

    composite_figure_t &operator=(const composite_figure_t &other)
    {
        auto tmp = other;
        std::swap(m_position, tmp.m_position);
        std::swap(m_figures, tmp.m_figures);
        return *this;
    }

    std::unique_ptr<figure_t> clone() const override
    {
        return std::make_unique<composite_figure_t>(*this);
    }

    void add(std::unique_ptr<figure_t> l)
    {
        m_figures.push_back(std::move(l));
    }

    double area() const override
    {
        double p = 0;
        for (const auto &figure : m_figures)
            p += figure->area();
        return p;
    }

    std::string_view class_name() const override
    {
        return "composite_figure_t";
    }

private:
    std::vector<std::unique_ptr<figure_t>> m_figures;
};

class print_observer_t : public figure_observer_t
{
public:
    void update(figure_t *who) override
    {
        std::cout << "Position of " << who << (*who) << " was updated\n";
    }
};

int main()
{
    rectangle_t r1(1, 2, 3, 4);

    r1.set_position({42, 6});

    //
    print_observer_t po;
    r1.add_observer(&po);

    r1.set_position({0, 0});
    r1.set_position({1, 1});

    //
    r1.remove_observer(&po);

    r1.set_position({42, 1});
}
